"""Test case generation for GeoVexEmbedder."""
import os
from pathlib import Path
from typing import Optional, cast

import geopandas as gpd
import h3
import torch
from h3ronpy.pandas.vector import cells_to_polygons
from pytorch_lightning import seed_everything

from srai.constants import REGIONS_INDEX, WGS84_CRS
from srai.embedders.geovex.embedder import GeoVexEmbedder
from srai.embedders.geovex.model import GeoVexModel
from srai.h3 import ring_buffer_h3_regions_gdf
from srai.joiners import IntersectionJoiner
from srai.loaders.osm_loaders import OSMPbfLoader
from srai.loaders.osm_loaders.filters import HEX2VEC_FILTER, OsmTagsFilter
from srai.neighbourhoods import H3Neighbourhood
from tests.embedders.geovex.constants import EMBEDDING_SIZE, TRAINER_KWARGS


def generate_test_case(
    test_case_name: str,
    root_region_index: str,
    region_gen_radius: int,
    model_radius: int,
    seed: int,
    tags: Optional[OsmTagsFilter] = None,
    convolutional_layer_size: int = 256,
    num_layers: int = 1,
) -> None:
    """Generate test case for GeoVexEmbedder."""
    seed_everything(seed, workers=True)
    os.environ["PYTHONHASHSEED"] = str(seed)
    torch.use_deterministic_algorithms(True)

    if tags is None:
        tags = HEX2VEC_FILTER

    neighbourhood = H3Neighbourhood(include_center=True)
    regions_indexes = list(
        neighbourhood.get_neighbours_up_to_distance(root_region_index, region_gen_radius)
    )

    geoms = cells_to_polygons([h3.str_to_int(r) for r in regions_indexes]).values
    regions_gdf = gpd.GeoDataFrame(index=regions_indexes, geometry=geoms, crs=WGS84_CRS)
    regions_gdf.index.name = REGIONS_INDEX

    regions_gdf = ring_buffer_h3_regions_gdf(regions_gdf, distance=model_radius).sort_index()
    buffered_geometry = regions_gdf.unary_union

    loader = OSMPbfLoader()
    features_gdf = loader.load(buffered_geometry, tags)

    joiner = IntersectionJoiner()
    joint_gdf = joiner.transform(regions_gdf, features_gdf)

    neighbourhood = H3Neighbourhood(regions_gdf)

    embedder = GeoVexEmbedder(
        target_features=[f"{st}_{t}" for st in tags for t in tags[st]],  # type: ignore
        batch_size=10,
        neighbourhood_radius=model_radius,
        embedding_size=EMBEDDING_SIZE,
        convolutional_layers=num_layers,
        convolutional_layer_size=convolutional_layer_size,
    )

    counts_df, _, _ = embedder._prepare_dataset(
        regions_gdf, features_gdf, joint_gdf, neighbourhood, embedder._batch_size, shuffle=True
    )

    embedder._prepare_model(counts_df, 0.001)

    for _, param in cast(GeoVexModel, embedder._model).named_parameters():
        param.data.fill_(0.01)

    results_df = embedder.fit_transform(
        regions_gdf=regions_gdf,
        features_gdf=features_gdf,
        joint_gdf=joint_gdf,
        neighbourhood=neighbourhood,
        trainer_kwargs=TRAINER_KWARGS,
        learning_rate=0.001,
    )

    results_df.columns = results_df.columns.astype(str)

    files_prefix = f"{test_case_name}"

    output_path = Path(__file__).parent / "test_files"
    regions_gdf.to_parquet(
        output_path / f"{files_prefix}_regions.parquet",
    )
    features_gdf.to_parquet(output_path / f"{files_prefix}_features.parquet", compression="gzip")
    joint_gdf.to_parquet(output_path / f"{files_prefix}_joint.parquet")
    results_df.to_parquet(output_path / f"{files_prefix}_result.parquet")


def generate_test_case_batches(
    test_case_name: str,
    root_region_index: str,
    region_gen_radius: int,
    model_radius: int,
    seed: int,
    tags: Optional[OsmTagsFilter] = None,
    convolutional_layer_size: int = 256,
    num_layers: int = 1,
) -> None:
    """Generate test case for GeoVexEmbedder."""
    seed_everything(seed, workers=True)
    os.environ["PYTHONHASHSEED"] = str(seed)
    torch.use_deterministic_algorithms(True)

    if tags is None:
        tags = HEX2VEC_FILTER

    neighbourhood = H3Neighbourhood(include_center=True)
    regions_indexes = list(
        neighbourhood.get_neighbours_up_to_distance(root_region_index, region_gen_radius)
    )

    geoms = cells_to_polygons([h3.str_to_int(r) for r in regions_indexes]).values
    regions_gdf = gpd.GeoDataFrame(index=regions_indexes, geometry=geoms, crs=WGS84_CRS)
    regions_gdf.index.name = REGIONS_INDEX

    regions_gdf = ring_buffer_h3_regions_gdf(regions_gdf, distance=model_radius).sort_index()
    buffered_geometry = regions_gdf.unary_union

    loader = OSMPbfLoader()
    features_gdf = loader.load(buffered_geometry, tags)

    joiner = IntersectionJoiner()
    joint_gdf = joiner.transform(regions_gdf, features_gdf)

    neighbourhood = H3Neighbourhood(regions_gdf)

    embedder = GeoVexEmbedder(
        target_features=[f"{st}_{t}" for st in tags for t in tags[st]],  # type: ignore
        batch_size=10,
        neighbourhood_radius=model_radius,
        embedding_size=EMBEDDING_SIZE,
        convolutional_layers=num_layers,
        convolutional_layer_size=convolutional_layer_size,
    )

    counts_df, dataloader, _ = embedder._prepare_dataset(
        regions_gdf, features_gdf, joint_gdf, neighbourhood, embedder._batch_size, shuffle=True
    )

    embedder._prepare_model(counts_df, 0.001)

    for _, param in cast(GeoVexModel, embedder._model).named_parameters():
        param.data.fill_(0.01)

    output_path = Path(__file__).parent / "test_files"
    files_prefix = f"{test_case_name}"

    for i, batch in enumerate(dataloader):
        torch.save(batch, output_path / f"{files_prefix}_batch_{i}.pt")

        encoder_forward_tensor = cast(GeoVexModel, embedder._model).encoder.forward(batch)
        torch.save(encoder_forward_tensor, output_path / f"{files_prefix}_encoder_forward_{i}.pt")

        forward_tensor = cast(GeoVexModel, embedder._model).forward(batch)
        torch.save(forward_tensor, output_path / f"{files_prefix}_forward_{i}.pt")

        loss_tensor = cast(GeoVexModel, embedder._model).training_step(batch, i)
        torch.save(loss_tensor, output_path / f"{files_prefix}_loss_{i}.pt")


if __name__ == "__main__":
    from constants import PREDEFINED_TEST_CASES

    for test_case in PREDEFINED_TEST_CASES:
        generate_test_case(**test_case)
        generate_test_case_batches(**test_case)



"""Neighbourhood interface."""

from abc import ABC, abstractmethod
from queue import Queue
from typing import Generic, Optional, TypeVar

from functional import seq

IndexType = TypeVar("IndexType")


class Neighbourhood(ABC, Generic[IndexType]):
    """
    Neighbourhood interface.

    This class abstracts away getting the neighbours of a region. It allows to get the neighbours at
    a certain distance or up to a certain distance. It is worth noting, that the distance here is
    not a metric distance, but a number of hops. This definition makes most sense semantically for
    grid systems such as H3 or S2 but should work for arbitrary neighbourhoods as well.

    The subclasses only need to implement the `get_neighbours` method, but can also override the
    `get_neighbours_up_to_distance` and `get_neighbours_at_distance` methods for performance
    reasons.  See the `H3Neighbourhood` class for an example. The class also provides a
    `_handle_center` method, which can be used to handle including/excluding the center region.
    """

    def __init__(self, include_center: bool = False) -> None:
        """
        Initializes the Neighbourhood.

        Args:
            include_center (bool): Whether to include the region itself in the neighbours.
            This is the default value used for all the methods of the class,
            unless overridden in the function call.
        """
        super().__init__()
        self.include_center = include_center

    @abstractmethod
    def get_neighbours(
        self, index: IndexType, include_center: Optional[bool] = None
    ) -> set[IndexType]:
        """
        Get the direct neighbours of a region using its index.

        Args:
            index (IndexType): Unique identifier of the region.
                Dependant on the implementation.
            include_center (Optional[bool]): Whether to include the region itself in the neighbours.
            If None, the value set in __init__ is used. Defaults to None.

        Returns:
            Set[IndexType]: Indexes of the neighbours.
        """

    def get_neighbours_up_to_distance(
        self, index: IndexType, distance: int, include_center: Optional[bool] = None
    ) -> set[IndexType]:
        """
        Get the neighbours of a region up to a certain distance.

        Args:
            index (IndexType): Unique identifier of the region.
                Dependant on the implementation.
            distance (int): Maximum distance to the neighbours.
            include_center (Optional[bool]): Whether to include the region itself in the neighbours.
            If None, the value set in __init__ is used. Defaults to None.

        Returns:
            Set[IndexType]: Indexes of the neighbours.
        """
        neighbours_with_distances = self._get_neighbours_with_distances(index, distance)
        neighbours: set[IndexType] = seq(neighbours_with_distances).map(lambda x: x[0]).to_set()
        neighbours = self._handle_center(
            index, distance, neighbours, at_distance=False, include_center_override=include_center
        )
        return neighbours

    def get_neighbours_at_distance(
        self, index: IndexType, distance: int, include_center: Optional[bool] = None
    ) -> set[IndexType]:
        """
        Get the neighbours of a region at a certain distance.

        Args:
            index (IndexType): Unique identifier of the region.
                Dependant on the implementation.
            distance (int): Distance to the neighbours.
            include_center (Optional[bool]): Whether to include the region itself in the neighbours.
            If None, the value set in __init__ is used. Defaults to None.

        Returns:
            Set[IndexType]: Indexes of the neighbours.
        """
        neighbours_up_to_distance = self._get_neighbours_with_distances(index, distance)
        neighbours_at_distance: set[IndexType] = (
            seq(neighbours_up_to_distance)
            .filter(lambda x: x[1] == distance)
            .map(lambda x: x[0])
            .to_set()
        )
        neighbours_at_distance = self._handle_center(
            index,
            distance,
            neighbours_at_distance,
            at_distance=True,
            include_center_override=include_center,
        )
        return neighbours_at_distance

    def _get_neighbours_with_distances(
        self, index: IndexType, distance: int
    ) -> set[tuple[IndexType, int]]:
        visited_indexes: dict[IndexType, int] = {}
        to_visit: Queue[tuple[IndexType, int]] = Queue()

        to_visit.put((index, 0))
        while not to_visit.empty():
            current_index, current_distance = to_visit.get()

            visited_indexes[current_index] = min(
                current_distance, visited_indexes.get(current_index, distance)
            )
            if current_distance < distance:
                current_neighbours = self.get_neighbours(current_index)
                for neighbour in current_neighbours:
                    if neighbour not in visited_indexes:
                        to_visit.put((neighbour, current_distance + 1))

        return set(visited_indexes.items())

    def _handle_center(
        self,
        index: IndexType,
        distance: int,
        neighbours: set[IndexType],
        at_distance: bool,
        include_center_override: Optional[bool],
    ) -> set[IndexType]:
        if include_center_override is None:
            include_center = self.include_center
        else:
            include_center = include_center_override

        if distance < 0:
            return set()
        elif distance == 0:
            if include_center:
                neighbours.add(index)
            else:
                neighbours.discard(index)
        else:
            if at_distance:
                neighbours.discard(index)
            else:
                if include_center:
                    neighbours.add(index)
                else:
                    neighbours.discard(index)
        return neighbours



"""CountEmbedder tests."""
from contextlib import nullcontext as does_not_raise
from typing import TYPE_CHECKING, Any, List, Union
from unittest import TestCase

import pandas as pd
import pytest
from pandas.testing import assert_frame_equal

from srai.constants import REGIONS_INDEX
from srai.embedders import CountEmbedder
from srai.loaders.osm_loaders.filters import GroupedOsmTagsFilter, OsmTagsFilter

if TYPE_CHECKING:  # pragma: no cover
    import geopandas as gpd

ut = TestCase()


@pytest.fixture  # type: ignore
def expected_embedding_df() -> pd.DataFrame:
    """Get expected CountEmbedder output for the default case."""
    expected_df = pd.DataFrame(
        {
            REGIONS_INDEX: ["891e2040897ffff", "891e2040d4bffff", "891e2040d5bffff"],
            "leisure": [0, 1, 1],
            "amenity": [1, 0, 1],
        },
    )
    expected_df.set_index(REGIONS_INDEX, inplace=True)

    return expected_df


@pytest.fixture  # type: ignore
def expected_subcategories_embedding_df() -> pd.DataFrame:
    """Get expected CountEmbedder output with subcategories for the default case."""
    expected_df = pd.DataFrame(
        {
            REGIONS_INDEX: ["891e2040897ffff", "891e2040d4bffff", "891e2040d5bffff"],
            "leisure_adult_gaming_centre": [0, 0, 1],
            "leisure_playground": [0, 1, 0],
            "amenity_pub": [1, 0, 1],
        },
    )
    expected_df.set_index(REGIONS_INDEX, inplace=True)

    return expected_df


@pytest.fixture  # type: ignore
def expected_feature_names() -> List[str]:
    """Get expected feature names for CountEmbedder."""
    expected_feature_names = ["amenity_parking", "leisure_park", "amenity_pub"]
    return expected_feature_names


@pytest.fixture  # type: ignore
def osm_tags_filter() -> OsmTagsFilter:
    """Get osm tags filter for CountEmbedder."""
    tags_filter: OsmTagsFilter = {
        "amenity": ["parking", "pub"],
        "leisure": "park",
    }
    return tags_filter


@pytest.fixture  # type: ignore
def specified_features_expected_embedding_df() -> pd.DataFrame:
    """Get expected CountEmbedder output for the case with specified features."""
    expected_df = pd.DataFrame(
        {
            REGIONS_INDEX: ["891e2040897ffff", "891e2040d4bffff", "891e2040d5bffff"],
            "amenity_parking": [0, 0, 0],
            "leisure_park": [0, 0, 0],
            "amenity_pub": [0, 0, 0],
        },
    )
    expected_df.set_index(REGIONS_INDEX, inplace=True)

    return expected_df


@pytest.fixture  # type: ignore
def specified_subcategories_features_expected_embedding_df() -> pd.DataFrame:
    """Get expected CountEmbedder output with subcategories for the case with specified features."""
    expected_df = pd.DataFrame(
        {
            REGIONS_INDEX: ["891e2040897ffff", "891e2040d4bffff", "891e2040d5bffff"],
            "amenity_parking": [0, 0, 0],
            "leisure_park": [0, 0, 0],
            "amenity_pub": [1, 0, 1],
        },
    )
    expected_df.set_index(REGIONS_INDEX, inplace=True)

    return expected_df


@pytest.mark.parametrize(  # type: ignore
    "regions_fixture,features_fixture,joint_fixture,expected_embedding_fixture,count_subcategories,expected_features_fixture",
    [
        (
            "gdf_regions",
            "gdf_features",
            "gdf_joint",
            "expected_embedding_df",
            False,
            None,
        ),
        (
            "gdf_regions",
            "gdf_features",
            "gdf_joint",
            "expected_subcategories_embedding_df",
            True,
            None,
        ),
        (
            "gdf_regions",
            "gdf_features",
            "gdf_joint",
            "specified_features_expected_embedding_df",
            False,
            "expected_feature_names",
        ),
        (
            "gdf_regions",
            "gdf_features",
            "gdf_joint",
            "specified_subcategories_features_expected_embedding_df",
            True,
            "expected_feature_names",
        ),
        (
            "gdf_regions",
            "gdf_features",
            "gdf_joint",
            "expected_embedding_df",
            False,
            "osm_tags_filter",
        ),
        (
            "gdf_regions",
            "gdf_features",
            "gdf_joint",
            "specified_subcategories_features_expected_embedding_df",
            True,
            "osm_tags_filter",
        ),
    ],
)
def test_correct_embedding(
    regions_fixture: str,
    features_fixture: str,
    joint_fixture: str,
    expected_embedding_fixture: str,
    count_subcategories: bool,
    expected_features_fixture: Union[str, None],
    request: Any,
) -> None:
    """Test if CountEmbedder returns correct result with different parameters."""
    expected_output_features = (
        None
        if expected_features_fixture is None
        else request.getfixturevalue(expected_features_fixture)
    )
    embedder = CountEmbedder(
        expected_output_features=expected_output_features, count_subcategories=count_subcategories
    )
    gdf_regions: gpd.GeoDataFrame = request.getfixturevalue(regions_fixture)
    gdf_features: gpd.GeoDataFrame = request.getfixturevalue(features_fixture)
    gdf_joint: gpd.GeoDataFrame = request.getfixturevalue(joint_fixture)
    embedding_df = embedder.transform(
        regions_gdf=gdf_regions, features_gdf=gdf_features, joint_gdf=gdf_joint
    )
    expected_result_df = request.getfixturevalue(expected_embedding_fixture)
    print(expected_embedding_fixture)
    print(expected_result_df)
    print(embedding_df)
    assert_frame_equal(
        embedding_df.sort_index(axis=1), expected_result_df.sort_index(axis=1), check_dtype=False
    )


@pytest.mark.parametrize(  # type: ignore
    "regions_fixture,features_fixture,joint_fixture,expected_features_fixture,expectation",
    [
        (
            "gdf_regions_empty",
            "gdf_features",
            "gdf_joint",
            None,
            does_not_raise(),
        ),
        (
            "gdf_regions",
            "gdf_features_empty",
            "gdf_joint",
            None,
            pytest.raises(ValueError),
        ),
        (
            "gdf_regions",
            "gdf_features_empty",
            "gdf_joint",
            "expected_feature_names",
            does_not_raise(),
        ),
        (
            "gdf_regions",
            "gdf_features",
            "gdf_joint_empty",
            None,
            does_not_raise(),
        ),
    ],
)
def test_empty(
    regions_fixture: str,
    features_fixture: str,
    joint_fixture: str,
    expected_features_fixture: Union[str, None],
    expectation: Any,
    request: Any,
) -> None:
    """Test CountEmbedder handling of empty input data frames."""
    expected_output_features = (
        None
        if expected_features_fixture is None
        else request.getfixturevalue(expected_features_fixture)
    )
    embedder = CountEmbedder(expected_output_features)
    gdf_regions: gpd.GeoDataFrame = request.getfixturevalue(regions_fixture)
    gdf_features: gpd.GeoDataFrame = request.getfixturevalue(features_fixture)
    gdf_joint: gpd.GeoDataFrame = request.getfixturevalue(joint_fixture)

    with expectation:
        embedding = embedder.transform(gdf_regions, gdf_features, gdf_joint)
        assert len(embedding) == len(gdf_regions)
        assert embedding.index.name == gdf_regions.index.name
        if expected_output_features:
            assert embedding.columns.tolist() == expected_output_features

        assert (embedding == 0).all().all()


@pytest.mark.parametrize(  # type: ignore
    "regions_fixture,features_fixture,joint_fixture,expectation",
    [
        (
            "gdf_unnamed_single_index",
            "gdf_features",
            "gdf_joint",
            pytest.raises(ValueError),
        ),
        (
            "gdf_regions",
            "gdf_unnamed_single_index",
            "gdf_joint",
            pytest.raises(ValueError),
        ),
        (
            "gdf_regions",
            "gdf_features",
            "gdf_unnamed_single_index",
            pytest.raises(ValueError),
        ),
        (
            "gdf_regions",
            "gdf_features",
            "gdf_three_level_multi_index",
            pytest.raises(ValueError),
        ),
        (
            "gdf_incorrectly_named_single_index",
            "gdf_features",
            "gdf_joint",
            pytest.raises(ValueError),
        ),
        (
            "gdf_regions",
            "gdf_incorrectly_named_single_index",
            "gdf_joint",
            pytest.raises(ValueError),
        ),
    ],
)
def test_incorrect_indexes(
    regions_fixture: str, features_fixture: str, joint_fixture: str, expectation: Any, request: Any
) -> None:
    """Test if cannot embed with incorrect dataframe indexes."""
    regions_gdf = request.getfixturevalue(regions_fixture)
    features_gdf = request.getfixturevalue(features_fixture)
    joint_gdf = request.getfixturevalue(joint_fixture)

    with expectation:
        CountEmbedder().transform(
            regions_gdf=regions_gdf, features_gdf=features_gdf, joint_gdf=joint_gdf
        )


@pytest.mark.parametrize(  # type: ignore
    "osm_tags_filter,count_subcategories,expected_output_features,expectation",
    [
        (
            {"amenity": True},
            True,
            None,
            pytest.raises(ValueError),
        ),
        (
            {"amenity": True},
            False,
            ["amenity"],
            does_not_raise(),
        ),
        (
            {"amenity": "pub"},
            True,
            ["amenity_pub"],
            does_not_raise(),
        ),
        (
            {"amenity": ["pub", "parking"]},
            True,
            ["amenity_pub", "amenity_parking"],
            does_not_raise(),
        ),
        (
            {"amenity": "pub"},
            False,
            ["amenity"],
            does_not_raise(),
        ),
        (
            {"amenity": ["pub", "parking"]},
            False,
            ["amenity"],
            does_not_raise(),
        ),
        (
            {"amenity": ["pub", "parking"], "leisure": ["park"]},
            True,
            ["amenity_pub", "amenity_parking", "leisure_park"],
            does_not_raise(),
        ),
        (
            {"amenity": ["pub", "parking"], "leisure": ["park"]},
            False,
            ["amenity", "leisure"],
            does_not_raise(),
        ),
        (
            {"group": {"amenity": True}},
            True,
            None,
            pytest.raises(ValueError),
        ),
        (
            {"group": {"amenity": True}},
            False,
            ["group"],
            does_not_raise(),
        ),
        (
            {"group": {"amenity": "pub"}},
            True,
            ["group_amenity=pub"],
            does_not_raise(),
        ),
        (
            {"group": {"amenity": ["pub", "parking"]}},
            True,
            ["group_amenity=pub", "group_amenity=parking"],
            does_not_raise(),
        ),
        (
            {"group": {"amenity": "pub"}},
            False,
            ["group"],
            does_not_raise(),
        ),
        (
            {"group": {"amenity": ["pub", "parking"]}},
            False,
            ["group"],
            does_not_raise(),
        ),
        (
            {"group": {"amenity": ["pub", "parking"], "leisure": ["park"]}},
            True,
            ["group_amenity=pub", "group_amenity=parking", "group_leisure=park"],
            does_not_raise(),
        ),
        (
            {"group": {"amenity": ["pub", "parking"], "leisure": ["park"]}},
            False,
            ["group"],
            does_not_raise(),
        ),
    ],
)
def test_osm_tags_filter_parsing(
    osm_tags_filter: Union[OsmTagsFilter, GroupedOsmTagsFilter],
    count_subcategories: bool,
    expected_output_features: str,
    expectation: Any,
) -> None:
    """Test is properly parses osm tags filter."""
    with expectation:
        embedder = CountEmbedder(
            expected_output_features=osm_tags_filter, count_subcategories=count_subcategories
        )

        ut.assertCountEqual(embedder.expected_output_features, expected_output_features)



"""Highway2VecEmbedder tests."""
import geopandas as gpd
import numpy as np
import pandas as pd
import pytest
from pytorch_lightning import seed_everything
from shapely.geometry import LineString, Polygon

from srai.constants import FEATURES_INDEX, REGIONS_INDEX
from srai.embedders import Highway2VecEmbedder
from srai.exceptions import ModelNotFitException


@pytest.fixture  # type: ignore
def highway2vec_features() -> gpd.GeoDataFrame:
    """Get Highway2Vec features GeoDataFrame."""
    features_gdf = gpd.GeoDataFrame(
        {
            "oneway": [1, 0, 0],
            "lanes-1": [1, 1, 0],
            "lanes-2": [0, 0, 1],
            "bicycle-destination": [0, 0, 0],
        },
        geometry=[
            LineString([(1, 1), (2, 2), (5, 5)]),
            LineString([(1, 2), (2, 3), (3, 4)]),
            LineString([(5, 5), (6, 1), (10, 2)]),
        ],
        index=[1, 2, 3],
    )
    features_gdf.index.name = FEATURES_INDEX
    return features_gdf


@pytest.fixture  # type: ignore
def highway2vec_regions() -> gpd.GeoDataFrame:
    """Get Highway2Vec regions GeoDataFrame."""
    regions_gdf = gpd.GeoDataFrame(
        {
            REGIONS_INDEX: ["ff1", "ff2", "ff3"],
        },
        geometry=[
            Polygon([(0, 0), (0, 3), (3, 3), (3, 0)]),
            Polygon([(4, 0), (4, 3), (7, 3), (7, 0)]),
            Polygon([(8, 0), (8, 3), (11, 3), (11, 0)]),
        ],
    ).set_index(REGIONS_INDEX)
    return regions_gdf


@pytest.fixture  # type: ignore
def highway2vec_joint() -> gpd.GeoDataFrame:
    """Get Highway2Vec joint GeoDataFrame."""
    joint_gdf = gpd.GeoDataFrame()
    joint_gdf.index = pd.MultiIndex.from_tuples(
        [("ff1", 1), ("ff1", 2), ("ff2", 3), ("ff3", 3)],
        names=[REGIONS_INDEX, FEATURES_INDEX],
    )
    return joint_gdf


@pytest.fixture  # type: ignore
def highway2vec_embeddings() -> pd.DataFrame:
    """Get features embedded."""
    embeddings = np.array(
        [
            [-0.019932, 0.027169, -0.031977, -0.000582],
            [0.236779, -0.258233, 0.051689, 0.098861],
            [0.236779, -0.258233, 0.051689, 0.098861],
        ],
        dtype=np.float32,
    )

    features = pd.DataFrame(
        embeddings,
        index=pd.Index(["ff1", "ff2", "ff3"], name=REGIONS_INDEX),
        columns=pd.RangeIndex(0, 4, 1),
    )
    return features


def test_transform_with_unfit_model(
    highway2vec_regions: gpd.GeoDataFrame,
    highway2vec_features: gpd.GeoDataFrame,
    highway2vec_joint: gpd.GeoDataFrame,
) -> None:
    """Test Highway2VecEmbedder transform with unfitted model."""
    embedder = Highway2VecEmbedder()
    with pytest.raises(ModelNotFitException):
        embedder.transform(highway2vec_regions, highway2vec_features, highway2vec_joint)


def test_embedder_on_correct_input(
    highway2vec_regions: gpd.GeoDataFrame,
    highway2vec_features: gpd.GeoDataFrame,
    highway2vec_joint: gpd.GeoDataFrame,
    highway2vec_embeddings: gpd.GeoDataFrame,
) -> None:
    """Test Highway2VecEmbedder results."""
    embedder = Highway2VecEmbedder(embedding_size=4)

    seed_everything(42)
    embedder.fit(highway2vec_regions, highway2vec_features, highway2vec_joint)
    features_embedded = embedder.transform(
        highway2vec_regions, highway2vec_features, highway2vec_joint
    )
    pd.testing.assert_frame_equal(features_embedded, highway2vec_embeddings, atol=1e-3)

    seed_everything(42)
    features_embedded = embedder.fit_transform(
        highway2vec_regions, highway2vec_features, highway2vec_joint
    )
    pd.testing.assert_frame_equal(features_embedded, highway2vec_embeddings, atol=1e-3)



"""Module contains a dedicated type alias for OSM tags filter."""
from typing import Dict, List, Union, cast

from srai._typing import is_expected_type

OsmTagsFilter = Dict[str, Union[List[str], str, bool]]

GroupedOsmTagsFilter = Dict[str, OsmTagsFilter]


def merge_grouped_osm_tags_filter(grouped_filter: GroupedOsmTagsFilter) -> OsmTagsFilter:
    """
    Merge grouped osm tags filter into a base one.

    Function merges all filter categories into a single one for an OSM loader to use.

    Args:
        grouped_filter (GroupedOsmTagsFilter): Grouped filter to be merged into a single one.

    Returns:
        osm_tags_type: Merged filter.
    """
    if not is_expected_type(grouped_filter, GroupedOsmTagsFilter):
        raise ValueError(
            "Provided filter doesn't match required `GroupedOsmTagsFilter` definition."
        )

    result: OsmTagsFilter = {}
    for sub_filter in grouped_filter.values():
        for osm_tag_key, osm_tag_value in sub_filter.items():
            if osm_tag_key not in result:
                result[osm_tag_key] = []

            # If filter is already a positive boolean, skip
            if isinstance(result[osm_tag_key], bool) and result[osm_tag_key]:
                continue

            # Check bool
            if isinstance(osm_tag_value, bool) and osm_tag_value:
                result[osm_tag_key] = True
            # Check string
            elif isinstance(osm_tag_value, str) and osm_tag_value not in cast(
                List[str], result[osm_tag_key]
            ):
                cast(List[str], result[osm_tag_key]).append(osm_tag_value)
            # Check list
            elif isinstance(osm_tag_value, list):
                new_values = [
                    value
                    for value in osm_tag_value
                    if value not in cast(List[str], result[osm_tag_key])
                ]
                cast(List[str], result[osm_tag_key]).extend(new_values)

    return result



"""Tests for neigbourhoods with regions."""
from typing import Any, Set

import geopandas as gpd
import pytest

from srai.neighbourhoods import H3Neighbourhood


@pytest.fixture  # type: ignore
def empty_gdf() -> gpd.GeoDataFrame:
    """Fixture for an empty GeoDataFrame."""
    return gpd.GeoDataFrame()


@pytest.fixture  # type: ignore
def single_hex_gdf() -> gpd.GeoDataFrame:
    """Fixture for a GeoDataFrame with a single hexagon."""
    return gpd.GeoDataFrame(
        index=["811e3ffffffffff"],
    )


@pytest.fixture  # type: ignore
def hex_without_one_neighbour_gdf() -> gpd.GeoDataFrame:
    """Fixture for a GeoDataFrame with a single hexagon, and its neighbours except for one."""
    return gpd.GeoDataFrame(
        index=[
            "811e3ffffffffff",
            "811f3ffffffffff",
            "811fbffffffffff",
            "811ebffffffffff",
            "811efffffffffff",
            "811e7ffffffffff",
        ],
    )


@pytest.fixture  # type: ignore
def two_rings_regions_some_missing() -> gpd.GeoDataFrame:
    """
    Fixture for a GeoDataFrame with regions up to two rings around initial one.

    Some regions are missing.
    """
    initial_region = ["862bac507ffffff"]
    first_ring = [
        "862bac50fffffff",
        "862bac517ffffff",
        "862bac51fffffff",
        "862bac527ffffff",
        "862bac52fffffff",
        "862bac537ffffff",
    ]
    second_ring = [
        "862ba124fffffff",
        "862ba126fffffff",
        "862bac427ffffff",
        "862bac437ffffff",
        "862bac557ffffff",
        "862bac577ffffff",
        "862bac5a7ffffff",
        "862bac5afffffff",
        "862bacc8fffffff",
        "862bacc9fffffff",
        "862baccd7ffffff",
        "862baccdfffffff",
    ]
    # This is the result of running h3.grid_disk("862bac507ffffff", 2)
    indices = initial_region + first_ring + second_ring

    # Explicitly remove the 'missing' regions so it's clear what's going on
    # Remove two regions from first ring
    indices.remove("862bac52fffffff")
    indices.remove("862bac537ffffff")
    # Remove three regions from second ring
    indices.remove("862bacc9fffffff")
    indices.remove("862baccd7ffffff")
    indices.remove("862baccdfffffff")
    return gpd.GeoDataFrame(index=indices)


@pytest.mark.parametrize(  # type: ignore
    "regions_gdf_fixture,expected,expected_with_include_center",
    [
        (
            "empty_gdf",
            set(),
            set(),
        ),
        (
            "single_hex_gdf",
            set(),
            {"811e3ffffffffff"},
        ),
        (
            "hex_without_one_neighbour_gdf",
            {
                "811f3ffffffffff",
                "811fbffffffffff",
                "811ebffffffffff",
                "811efffffffffff",
                "811e7ffffffffff",
            },
            {
                "811f3ffffffffff",
                "811fbffffffffff",
                "811ebffffffffff",
                "811efffffffffff",
                "811e7ffffffffff",
                "811e3ffffffffff",
            },
        ),
    ],
)
def test_get_neighbours_with_regions_gdf(
    regions_gdf_fixture: str,
    expected: Set[str],
    expected_with_include_center: Set[str],
    request: Any,
) -> None:
    """Test get_neighbours of H3Neighbourhood with a specified regions GeoDataFrame."""
    center_region = "811e3ffffffffff"
    regions_gdf = request.getfixturevalue(regions_gdf_fixture)

    neighbourhood = H3Neighbourhood(regions_gdf)
    assert neighbourhood.get_neighbours(center_region) == expected
    assert (
        neighbourhood.get_neighbours(center_region, include_center=True)
        == expected_with_include_center
    )

    neighbourhood_with_include_center = H3Neighbourhood(regions_gdf, include_center=True)
    assert (
        neighbourhood_with_include_center.get_neighbours(center_region)
        == expected_with_include_center
    )
    assert (
        neighbourhood_with_include_center.get_neighbours(center_region, include_center=False)
        == expected
    )


@pytest.mark.parametrize(  # type: ignore
    "distance,expected,expected_with_include_center",
    [
        (
            0,
            set(),
            {"862bac507ffffff"},
        ),
        (
            1,
            {
                "862bac50fffffff",
                "862bac517ffffff",
                "862bac51fffffff",
                "862bac527ffffff",
            },
            {
                "862bac50fffffff",
                "862bac517ffffff",
                "862bac51fffffff",
                "862bac527ffffff",
                "862bac507ffffff",
            },
        ),
        (
            2,
            {
                "862bac50fffffff",
                "862bac517ffffff",
                "862bac51fffffff",
                "862bac527ffffff",
                "862ba124fffffff",
                "862ba126fffffff",
                "862bac427ffffff",
                "862bac437ffffff",
                "862bac557ffffff",
                "862bac577ffffff",
                "862bac5a7ffffff",
                "862bac5afffffff",
                "862bacc8fffffff",
            },
            {
                "862bac50fffffff",
                "862bac517ffffff",
                "862bac51fffffff",
                "862bac527ffffff",
                "862ba124fffffff",
                "862ba126fffffff",
                "862bac427ffffff",
                "862bac437ffffff",
                "862bac557ffffff",
                "862bac577ffffff",
                "862bac5a7ffffff",
                "862bac5afffffff",
                "862bacc8fffffff",
                "862bac507ffffff",
            },
        ),
    ],
)
def test_get_neighbours_up_to_distance_with_regions_gdf(
    distance: int, expected: Set[str], expected_with_include_center: Set[str], request: Any
) -> None:
    """Test get_neighbours_up_to_distance of H3Neighbourhood with a specified regions."""
    center_region = "862bac507ffffff"
    regions_gdf = request.getfixturevalue("two_rings_regions_some_missing")

    neighbourhood = H3Neighbourhood(regions_gdf)
    assert neighbourhood.get_neighbours_up_to_distance(center_region, distance) == expected
    assert (
        neighbourhood.get_neighbours_up_to_distance(center_region, distance, include_center=True)
        == expected_with_include_center
    )

    neighbourhood_with_include_center = H3Neighbourhood(regions_gdf, include_center=True)
    assert (
        neighbourhood_with_include_center.get_neighbours_up_to_distance(center_region, distance)
        == expected_with_include_center
    )
    assert (
        neighbourhood_with_include_center.get_neighbours_up_to_distance(
            center_region, distance, include_center=False
        )
        == expected
    )


@pytest.mark.parametrize(  # type: ignore
    "distance,expected,expected_with_include_center",
    [
        (
            0,
            set(),
            {"862bac507ffffff"},
        ),
        (
            1,
            {
                "862bac50fffffff",
                "862bac517ffffff",
                "862bac51fffffff",
                "862bac527ffffff",
            },
            {
                "862bac50fffffff",
                "862bac517ffffff",
                "862bac51fffffff",
                "862bac527ffffff",
            },
        ),
        (
            2,
            {
                "862ba124fffffff",
                "862ba126fffffff",
                "862bac427ffffff",
                "862bac437ffffff",
                "862bac557ffffff",
                "862bac577ffffff",
                "862bac5a7ffffff",
                "862bac5afffffff",
                "862bacc8fffffff",
            },
            {
                "862ba124fffffff",
                "862ba126fffffff",
                "862bac427ffffff",
                "862bac437ffffff",
                "862bac557ffffff",
                "862bac577ffffff",
                "862bac5a7ffffff",
                "862bac5afffffff",
                "862bacc8fffffff",
            },
        ),
    ],
)
def test_get_neighbours_at_distance_with_regions_gdf(
    distance: int, expected: Set[str], expected_with_include_center: Set[str], request: Any
) -> None:
    """Test get_neighbours_at_distance of H3Neighbourhood with a specified regions GeoDataFrame."""
    center_region = "862bac507ffffff"
    regions_gdf = request.getfixturevalue("two_rings_regions_some_missing")

    neighbourhood = H3Neighbourhood(regions_gdf)
    assert neighbourhood.get_neighbours_at_distance(center_region, distance) == expected
    assert (
        neighbourhood.get_neighbours_at_distance(center_region, distance, include_center=True)
        == expected_with_include_center
    )

    neighbourhood_with_include_center = H3Neighbourhood(regions_gdf, include_center=True)
    assert (
        neighbourhood_with_include_center.get_neighbours_at_distance(center_region, distance)
        == expected_with_include_center
    )
    assert (
        neighbourhood_with_include_center.get_neighbours_at_distance(
            center_region, distance, include_center=False
        )
        == expected
    )



"""This module contains classes of strategy for handling downloaded tiles."""
from abc import ABC, abstractmethod
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING, Any, Union

if TYPE_CHECKING:  # pragma: no cover
    from PIL import Image


class DataCollector(ABC):
    """Store collected images."""

    @abstractmethod
    def store(self, idx: str, data: "Image.Image") -> Any:
        """
        Apply action for object storage and returns data of it.

        Args:
            idx (str): id of tile
            data (Image.Image): tile
        """


class SavingDataCollector(DataCollector):
    """
    Save in disk collected images.

    Store paths.
    """

    def __init__(self, save_path: Union[str, Path], file_extension: str) -> None:
        """
        Initialize SavingDataCollector.

        Args:
            save_path (Union[str, Path]): root path for data
            file_extension (str): file name extension
        """
        super().__init__()
        if save_path is None or file_extension is None:
            raise ValueError
        self.save_path = Path(save_path)
        self.format = file_extension

    def store(self, idx: str, data: "Image.Image") -> Path:
        """
        Save image on disk. Returns path.

        Args:
            idx (str): id of tile
            data (Image.Image): tile
        """
        path = self.save_path / f"{idx}.{self.format}"
        data.save(path)
        return path


class InMemoryDataCollector(DataCollector):
    """Store data in object memory."""

    def __init__(self) -> None:
        """Initialize InMemoryDataCollector."""
        super().__init__()

    def store(self, idx: str, data: "Image.Image") -> "Image.Image":
        """
        Simply return object for usage.

        Args:
            idx (str): id of tile
            data (Image.Image): tile
        """
        return data


class DataCollectorType(str, Enum):
    """Define enums to choose one of known DataCollector implementations."""

    SAVE = "save"
    RETURN = "return"


def get_collector(collector_type: Union[DataCollectorType, str], **kwargs: Any) -> DataCollector:
    """
    Return DataCollector object of type specified by DataCollectorType enum.

    Args:
        collector_type (DataCollectorType): If SAVE returns SavingDataCollector.
        If RETURN returns InMemoryDataCollector.
        **kwargs (Any): Extra arguments used for SavingDataCollector object creation arguments.

    Returns:
        DataCollector: newly created object
    """
    if collector_type == DataCollectorType.SAVE:
        return SavingDataCollector(**kwargs)
    elif collector_type == DataCollectorType.RETURN:
        return InMemoryDataCollector()
    else:
        raise ValueError
